package article_test

import (
	"context"
	"fmt"
	"sync"
	"testing"
	"time"
)

func TestFireAndForget(t *testing.T) {
	ch := make(chan int)

	go func(ch chan int) {
		defer close(ch)

		for i := 0; i < 10; i++ {
			ch <- i
		}
	}(ch)

	var wg sync.WaitGroup
	wg.Add(1)

	go func(ch chan int) {
		defer wg.Done()

		for arg := range ch {
			t.Logf("%d\n", arg*2)
		}
	}(ch)

	wg.Wait()
}

func TestCallAndResponse(t *testing.T) {
	tasks := make(chan int)
	responses := make(chan string)

	var wg sync.WaitGroup
	wg.Add(2)

	go func(tasks chan int, response chan string) {
		defer close(tasks)
		defer wg.Done()

		for i := 0; i < 10; i++ {
			tasks <- i
			res := <-responses
			t.Logf("%s\n", res)
		}
	}(tasks, responses)

	go func(tasks chan int, responses chan string) {
		defer wg.Done()

		for arg := range tasks {
			responses <- fmt.Sprintf("%d", arg*2)
		}
	}(tasks, responses)

	wg.Wait()
	close(responses)
}

type Job struct {
	arg int
	ch  chan string
}

func (job *Job) Result() string {
	s := <-job.ch
	close(job.ch)
	return s
}

func (job *Job) Reply(s string) {
	job.ch <- s
}

func TestIndividualCallAndResponse(t *testing.T) {
	ctx, cancel := context.WithTimeout(context.Background(), time.Second*5)
	tasks := make(chan *Job)

	var wg sync.WaitGroup
	for i := 0; i < 10; i++ {
		wg.Add(1)
		go func(job *Job, tasks chan *Job) {
			defer wg.Done()

			tasks <- job

			res := job.Result()
			t.Logf("%s\n", res)
		}(&Job{ch: make(chan string), arg: i}, tasks)
	}

	go func(ctx context.Context, tasks chan *Job) {
		for {
			select {
			case job := <-tasks:
				job.Reply(fmt.Sprintf("%d", job.arg * 2))
			case <-ctx.Done():
				return
			}
		}
	}(ctx, tasks)

	wg.Wait()
	close(tasks)

	cancel()

}
